//
//  Transit.cpp
//  Tycho
//
//  Created by Michael McLaughlin on 9/26/21.
//
//  Code requires the AA+ library (not included in Supplementary Material)
//  Version used = 2.30
//  Ref: http://www.naughter.com/aa.html

#include "Transit.hpp"

Transit::Transit() {
   ULat = CAACoordinateTransformation::DMSToDegrees(55, 54, 16.79);  // GPS
   ULong = -CAACoordinateTransformation::DMSToDegrees(12, 41, 27.59);   // AA+ convention
   date = 2451545.0;  // J2000
}

void
Transit::getSolarTransit(long year, long month, long day, bool isGregorian) {
   date = CAADynamicalTime::UTC2TT(CAADate(year, month, day, isGregorian));
   double critT = 0, critAlt = 0;
   const double tol = 7.7e-7;   // 1 arc-second precision
   
   // Initialize
   transitAltitude = INFINITY;   // sentinel
   dblPair result = GSSearch(0, 1, tol);  // uses date offset
   critT = 0.5*(result.x + result.y);
   critAlt = getSolarAltitude(critT);
   transitAltitude = critAlt + CAARefraction::RefractionFromTrue(critAlt);    // degrees
   transitTime = critT + date;    // TT
}

double
Transit::getSolarAltitude(double time) {  // without refraction
   CAA2DCoordinate Topo;
   double jd = time + date;
   double LongtitudeAsHourAngle = CAACoordinateTransformation::DegreesToHours(ULong);
   double Long = CAASun::ApparentEclipticLongitude(jd, true);
   double Lat = CAASun::ApparentEclipticLatitude(jd, true);
   const CAA2DCoordinate Equatorial = CAACoordinateTransformation::Ecliptic2Equatorial(Long, Lat, CAANutation::TrueObliquityOfEcliptic(jd));
   double SunRad = CAAEarth::RadiusVector(jd, true);
   Topo = CAAParallax::Equatorial2Topocentric(Equatorial.X, Equatorial.Y, SunRad, ULong, ULat, 0, jd);
   double AST = CAASidereal::ApparentGreenwichSiderealTime(jd);
   double LocalHourAngle = AST - LongtitudeAsHourAngle - Topo.X;
   CAA2DCoordinate Horizontal = CAACoordinateTransformation::Equatorial2Horizontal(LocalHourAngle, Topo.Y, ULat);
   
   return Horizontal.Y;
}

// Golden-section search in [a, b] given a univariate myFunc
dblPair
Transit::GSSearch(double a, double b, double epsilon, bool maximize)
{
   return GSSearch2(a, b, epsilon, b - a, true, 0, 0, true, 0, 0, maximize);
}

// GSSearch helper (recursive)
static double invPhi = 0.5*(sqrt(5) - 1), invPhi2 = 0.5*(3 - sqrt(5));
dblPair
Transit::GSSearch2(double a, double b, double epsilon, double h,
          bool not_in_C, double c, double fc, bool not_in_D, double d, double fd, bool maximize)
{
   dblPair result = {a, b};
   
   if (fabs(h) <= epsilon)
      return result;
   if (not_in_C) {
      c = a + invPhi2*h;
      fc = getSolarAltitude(c);
   }
   if (not_in_D) {
      d = a + invPhi*h;
      fd = getSolarAltitude(d);
   }
   if (maximize) {
      if (fc < fd)
         result = GSSearch2(c, b, epsilon, invPhi*h, false, d, fd, true, 0, 0, maximize);
      else
         result = GSSearch2(a, d, epsilon, invPhi*h, true, 0, 0, false, c, fc, maximize);
   }
   else {   // minimization
      if (fc < fd)
         result = GSSearch2(a, d, epsilon, invPhi*h, true, 0, 0, false, c, fc, maximize);
      else
         result = GSSearch2(c, b, epsilon, invPhi*h, false, d, fd, true, 0, 0, maximize);
   }
  
   return result;
}
